// Namespace
var android = {};
android.selection = {};

android.selection.selectionStartRange = null;
android.selection.selectionEndRange = null;

/** The last point touched by the user. { 'x': xPoint, 'y': yPoint } */
android.selection.lastTouchPoint = null;

var realWidth = null;
var realHeight = null;
var initialLeft = null;
var initialTop = null;

android.selection.resetVars = function(){
	realWidth = null;
    android.selection.getReadiumFrame().contentWindow.onresize = function(event) {
        android.selection.resetVars();
    };
}

android.selection.initVars = function(){
	var rect = android.selection.getReadiumFrame().getBoundingClientRect();
	realWidth = rect.right - rect.left;
	realHeight = rect.bottom - rect.top;
	initialLeft = rect.left;
	initialTop = rect.top;
}

/**
 * Starts the touch and saves the given x and y coordinates as last touch point
 */
android.selection.startTouch = function(x, y) {
    if(!realWidth){
    	android.selection.initVars();
    }
    coeffW = android.selection.getReadiumFrame().contentWindow.innerWidth/realWidth;
    coeffH = android.selection.getReadiumFrame().contentWindow.innerHeight/realHeight;
    y = (y * coeffH + android.selection.getYOffset());
    x = (x * coeffW + android.selection.getXOffset());
    android.selection.lastTouchPoint = {'x': x, 'y': y};
};

/**
 *  Checks to see if there is a selection.
 *
 *  @return boolean
 */
android.selection.hasSelection = function() {
    return android.selection.getReadiumFrame().contentWindow.getSelection().toString().length > 0;
};

/**
 *  Clears the current selection.
 */
android.selection.clearSelection = function() {
    try {
        // if current selection clear it.
        var sel = android.selection.getReadiumFrame().contentWindow.getSelection();
        sel.removeAllRanges();
    }
    catch (e) {
        console.error("clearSelection" + e);
    }
};

/**
 *  Handles the long touch action by selecting the last touched element.
 */
android.selection.longTouch = function() {
    try {
        android.selection.clearSelection();
        var sel = android.selection.getReadiumFrame().contentWindow.getSelection();
        var range = android.selection.getReadiumFrame().contentWindow.document.caretRangeFromPoint(android.selection.lastTouchPoint.x, android.selection.lastTouchPoint.y);
        range.expand("word");
        var text = range.toString();
        if (text.length == 1) {
            var baseKind = jpntext.kind(text);
            if (baseKind != jpntext.KIND['ascii']) {
                try {
                    do {
                        range.setEnd(range.endContainer, range.endOffset + 1);
                        text = range.toString();
                        var kind = jpntext.kind(text);
                    } while (baseKind == kind);
                    range.setEnd(range.endContainer, range.endOffset - 1);
                }
                catch (e) {
                }
                try {
                    do {
                        range.setStart(range.startContainer, range.startOffset - 1);
                        text = range.toString();
                        var kind = jpntext.kind(text);
                    } while (baseKind == kind);
                    range.setStart(range.startContainer, range.startOffset + 1);
                }
                catch (e) {
                }
            }
        }
        if (text.length > 0) {
            sel.addRange(range);
            android.selection.saveSelectionStart();
            android.selection.saveSelectionEnd();
            android.selection.selectionChanged(true);
        }
     }
     catch (err) {
         console.error("longTouch" + err);
     }
};

/**
 * Tells the app to show the context menu.
 */
android.selection.selectionChanged = function(isReallyChanged) {
    try {


        var sel = android.selection.getReadiumFrame().contentWindow.getSelection();

        if (!sel) {
            return;
        }

        // Rangy
        var rangyRange = android.selection.getRange();

        // Text to send to the selection
        var text = android.selection.getReadiumFrame().contentWindow.getSelection().toString();
        var coord = android.selection.getSelectionCoords();
        coeffW = android.selection.getReadiumFrame().contentWindow.innerWidth/realWidth;
        coeffH = android.selection.getReadiumFrame().contentWindow.innerHeight/realHeight;
        var leftShift;
        var topShift;
		if(isAndroidVersionOverOrEqual(4,4,0)){
	        leftShift = -initialLeft;
	        topShift = -initialTop;
			rect = android.selection.getReadiumFrame().getBoundingClientRect();
            contentFrame = document.getElementById("scrolled-content-frame");
	        if(contentFrame != null){
	        	topShift = contentFrame.scrollTop;
	        }
		}else{
	        leftShift = android.selection.getXOffset();
	        topShift = android.selection.getYOffset();
		}
        var handleBounds = "{'left': " + (coord.x/coeffW - leftShift) + ", ";
        handleBounds += "'top': " + ((coord.y + coord.h)/coeffH - topShift) + ", ";
        handleBounds += "'right': " + (coord.x2/coeffW - leftShift) + ", ";
        handleBounds += "'bottom': " + ((coord.y2 + coord.h2)/coeffH - topShift) + "}";

        // Set the content width
        window.TextSelection.setContentWidth(android.selection.getReadiumFrame().contentWindow.document.body.clientWidth);

        // Tell the interface that the selection changed
        window.TextSelection.selectionChanged(rangyRange, text, handleBounds, coord.h, isReallyChanged);
    }
    catch (e) {
        console.error("selectionChanged" + e);
    }
};

android.selection.getRange = function() {
    var serializedRangeSelected = rangy.serializeSelection();
    var serializerModule = rangy.modules.Serializer;
    if (serializedRangeSelected != '') {
        if (rangy.supported && serializerModule && serializerModule.supported) {
            var beginingCurly = serializedRangeSelected.indexOf("{");
            serializedRangeSelected = serializedRangeSelected.substring(0, beginingCurly);
            return serializedRangeSelected;
        }
    }
}

/**
 * Returns the last touch point as a readable string.
 */
android.selection.lastTouchPointString = function(){
    if (android.selection.lastTouchPoint == null)
        return "undefined";
    return "{" + android.selection.lastTouchPoint.x + "," + android.selection.lastTouchPoint.y + "}";
};

android.selection.saveSelectionStart = function(){
    try {
        // Save the starting point of the selection
        var sel = android.selection.getReadiumFrame().contentWindow.getSelection();
        var range = sel.getRangeAt(0);
        var saveRange = android.selection.getReadiumFrame().contentWindow.document.createRange();
        saveRange.setStart(range.startContainer, range.startOffset);
        android.selection.selectionStartRange = saveRange;
    }
    catch (e) {
        console.error("saveSelectionStart" + e);
    }
};

android.selection.saveSelectionEnd = function(){
    try {
        // Save the end point of the selection
        var sel = android.selection.getReadiumFrame().contentWindow.getSelection();
        var range = sel.getRangeAt(0);
        var saveRange = android.selection.getReadiumFrame().contentWindow.document.createRange();
        saveRange.setStart(range.endContainer, range.endOffset);
        android.selection.selectionEndRange = saveRange;
    }
    catch (e) {
        console.error("saveSelectionEnd" + e);
    }
};

/**
 * Sets the last caret position for the start handle.
 */
android.selection.setStartPos = function(x, y){
    try {
        coeffW = android.selection.getReadiumFrame().contentWindow.innerWidth/realWidth;
        coeffH = android.selection.getReadiumFrame().contentWindow.innerHeight/realHeight;
        y = (y * coeffH + android.selection.getYOffset());
        x = (x * coeffW + android.selection.getXOffset());
        android.selection.selectBetweenHandles(android.selection.getReadiumFrame().contentWindow.document.caretRangeFromPoint(x, y), android.selection.selectionEndRange);
    }
    catch (e) {
        console.error("setStartPos" + e);
    }
};

/**
 * Sets the last caret position for the end handle.
 */
android.selection.setEndPos = function(x, y){
    try {
        coeffW = android.selection.getReadiumFrame().contentWindow.innerWidth/realWidth;
        coeffH = android.selection.getReadiumFrame().contentWindow.innerHeight/realHeight;
        y = (y * coeffH + android.selection.getYOffset());
        x = (x * coeffW + android.selection.getXOffset());
        android.selection.selectBetweenHandles(android.selection.selectionStartRange, android.selection.getReadiumFrame().contentWindow.document.caretRangeFromPoint(x, y));
    }
    catch (e) {
        console.error("setEndPos" + e);
    }
};

android.selection.restoreStartEndPos = function() {
    try {
        android.selection.selectBetweenHandles(android.selection.selectionEndRange, android.selection.selectionStartRange);
    }
    catch (e) {
        console.error("restoreStartEndPos" + e);
    }
};

/**
 *  Selects all content between the two handles
 */
android.selection.selectBetweenHandles = function(startCaret, endCaret) {
    try {
        if (startCaret && endCaret) {
            var rightOrder = startCaret.compareBoundaryPoints(Range.START_TO_END, endCaret) <= 0;
            if (rightOrder) {
                android.selection.selectionStartRange = startCaret;
                android.selection.selectionEndRange = endCaret;
            }
            else {
                startCaret = android.selection.selectionStartRange;
                endCaret = android.selection.selectionEndRange;
            }
            var range = android.selection.getReadiumFrame().contentWindow.document.createRange();
            range.setStart(startCaret.startContainer, startCaret.startOffset);
            range.setEnd(endCaret.startContainer, endCaret.startOffset);
            android.selection.clearSelection();
            var selection = android.selection.getReadiumFrame().contentWindow.getSelection();
            selection.addRange(range);
            android.selection.selectionChanged(rightOrder);
        }
        else {
            android.selection.selectionStartRange = startCaret;
            android.selection.selectionEndRange = endCaret;
        }
    }
    catch (e) {
        console.error("selectBetweenHandles" + e);
    }
};

function isAndroidVersionOverOrEqual(main, second, last) {
    var ua = navigator.userAgent;
    var match = ua.match(/Android\s([0-9\.]*)/);
    var version =  match ? match[1] : false;
    if(!version){
        return false;
    }else{
        var mainVer = parseInt(version.split(".")[0]);
        var secondVer = parseInt(version.split(".")[1]);
        var lastVer = parseInt(version.split(".")[2]);
        if(mainVer >= main){
            if(secondVer >= second){
                if(lastVer >= last){
                    return true;
                }
            }
        }
        return false;
    }
};

android.selection.getYOffset = function(){
	var rect = android.selection.getReadiumFrame().getBoundingClientRect();
	return -rect.top;
}

android.selection.getXOffset = function(){
	var rect = android.selection.getReadiumFrame().getBoundingClientRect();
	return -rect.left;
}

android.selection.getSelectionCoords = function(){
    var sel = android.selection.getReadiumFrame().contentWindow.document.selection, range;
    var x = 0, y = 0;
    var x2 = 0, y2 = 0;
    var h = 0, h2 = 0;
    if (android.selection.getReadiumFrame().contentWindow.getSelection) {
        sel = android.selection.getReadiumFrame().contentWindow.getSelection();
        if (sel.rangeCount) {
            range = sel.getRangeAt(0).cloneRange();
            if (range.getClientRects) {
                //range.collapse(true);
                var rectList = range.getClientRects();
                var rect = rectList[0];
                if(rect && rect.left && rect.top && rect.height){
                    x = rect.left;
                    y = rect.top;
                    h = rect.height;
                }
                var rect2 = rectList[rectList.length-1];
                if(rect2 && rect2.right && rect2.top && rect2.height){
                    x2 = rect2.right;
                    y2 = rect2.top;
                    h2 = rect2.height;
                }
            }
        }
    }
    return { x: x, y: y , x2:x2, y2:y2, h:h, h2:h2};
}

android.selection.getReadiumFrame = function(){
    var readiumFrame = document.getElementById("epubContentIframe");
    if(readiumFrame == null){
        readiumFrame = document.getElementsByTagName("iframe")[0];
        if(readiumFrame == null){
        }
    }
    //Test:
    //document.getElementsByTagName('body')[0].style.backgroundColor="red";
    //readiumFrame.contentWindow.document.getElementsByTagName('body')[0].style.backgroundColor="green";
    return readiumFrame;
}