/*
 Copyright (c) 2014 Publiwide
 http://www.publiwide.com

 Created: 10/14/14 10:17 AM
 Authors: Alexandru Popa, alexandru.popa@publiwide.com
 */

var publiwide = publiwide || {};

publiwide.annotations =
{
	version: "2.1",
    annotationAdded: new publiwide.utils.Listeners(),
    annotationRemoved: new publiwide.utils.Listeners,
    bookmarks: new utils.Dictionary(),
    highlights: new utils.Dictionary(),
    notes: new utils.Dictionary(),
    all: new utils.Dictionary(),

    _concatTextChildNodes: function(node)
    {
        var i = 0;
        while (i < node.childNodes.length - 1)
        {
            var child = node.childNodes.item(i);
            var nextChild = node.childNodes.item(i + 1);
            if (child.nodeType == 3 && nextChild.nodeType == 3)
            {
                child.nodeValue += nextChild.nodeValue;
                child.parentNode.removeChild(nextChild);
            }
            else
                i++;
            annotations._concatTextChildNodes(child);
        }
    },
    _getNoteIcon: function(note)
    {
        var icon = document.createElement('div');
        //icon.href = 'javascript:;';
        icon.id = 'aicon_' + note.key;
        icon.setAttribute('aid', note.key);
        icon.style.display = 'inline-block';
        icon.style.cursor = 'pointer';
        if (note.isPublic)
        {
            if (note.publicUserName == null || note.publicUserName.length == 0)
            {
                icon.title = 'Click to edit the note';
                icon.style.background = "url('" + utils.baseURL() + "images/annotations/public_note_icons.png') no-repeat";
                icon.style.width = '18px';
                icon.style.height = '21px';
                icon.style.position = 'relative';
                icon.style.top = '3px';
            }
            else
            {
                icon.title = 'Click to view the note created by ' + note.publicUserName;
                icon.style.background = "url('" + utils.baseURL() + "images/annotations/user_public_note_icon.png') no-repeat";
                icon.style.width = '17px';
                icon.style.height = '12px';
            }
        }
        else
        {
            icon.title = 'Click to edit the note';
            icon.style.background = "url('" + utils.baseURL() + "images/annotations/note_icons.png') no-repeat";
            icon.style.width = '11px';
            icon.style.height = '14px';
        }
        icon.style.marginRight = '3px';
        return icon;
    },
    _getNoteIconSVG: function(note)
    {
        var b = viewAnnot.getBody();
        var svg = b.getElementsByTagName('svg')[0];
        var icon = null;

        if (note.isPublic)
        {
            if (note.publicUserName == null || note.publicUserName.length == 0)
                icon = utils.createOn(svg, 'image', { 'opacity': '0.85', 'width': '18', 'height': '21', 'xlink:href': url + 'images/annotations/public_note' + note.category + '.png', 'aid': note.key });
            else
                icon = utils.createOn(svg, 'image', { 'opacity': '0.85', 'width': '17', 'height': '12', 'xlink:href': url + 'images/annotations/user_public_note_icon.png', 'aid': note.key });
        }
        else
            icon = utils.createOn(svg, 'image', { 'opacity': '0.85', 'width': '12', 'height': '14', 'xlink:href': ('images/annotations/note' + note.category + '.png'), 'aid': note.key });
        return icon;
    },
    _removeAnnotationTags: function(tag, id, attributeName)
    {
        var i = 0;
        try
        {
            while (i < tag.childNodes.length)
            {
                if (!annotations._removeAnnotationTags(tag.childNodes[i], id, attributeName))
                    i++;
            }
            //for (var i = 0; i < tag.childNodes.length; i++)
            //	annotations._removeAnnotationTags(tag.childNodes[i], id, attributeName);
            if (tag.attributes != null && tag.attributes[attributeName] != null && tag.attributes[attributeName].value == id && tag.parentNode != null)
            {
                var parent = tag.parentNode;
                for (var x = 0; x < tag.childNodes.length; x++)
                    parent.insertBefore(tag.childNodes[x], tag);
                parent.removeChild(tag);
                return true;
            }
            return false;
        }
        catch (error)
        {
            return false;
        }
    },
    _unifyLines: function(array)
    {
        if (array.length == 0)
            return array;
        var result = [array[0]];

        for (var i = 1; i < array.length; i++)
        {
            var rectI = array[i];
            var found = false;
            for (var j = 0; j < result.length; j++)
            {
                var rectJ = result[j];
                if (Math.abs(rectJ.top - rectI.top) <= 2 && Math.abs(rectJ.height - rectI.height) <= 2)
                {
                    var l = rectJ.left;
                    rectJ.left = Math.min(rectI.left, rectJ.left);
                    rectJ.top = Math.min(rectI.top, rectJ.top);
                    rectJ.height = Math.max(rectI.height, rectJ.height);
                    rectJ.width = Math.max(rectI.left + rectI.width, l + rectJ.width) - rectJ.left;
                    found = true;
                    break;
                }
            }
            if (!found)
                result.push(rectI);
        }

        return result;
    },

    addBookmark: function(bookmark)
    {
        var b = null;
		if (typeof bookmark.id != 'undefined' && bookmark.id > 0)
		{
			b = typeof annotations.bookmarks[bookmark.id] != 'undefined' ? annotations.bookmarks[bookmark.id] : new Bookmark();
			b.copyFrom(bookmark);
			annotations.bookmarks[bookmark.id] = b;
		}
		else
		{
			b = new Bookmark();
			b.copyFrom(bookmark);
			b.id = 0;
		}
		if (b != null)
			annotations.all[b.key] = b;
		return b;
    },
    addHighlight: function(highlight)
    {
        if (highlight == null || highlight.deleted)
            return null;
        var h = null;
        if (typeof highlight.id != 'undefined' && highlight.id > 0)
        {
            h = typeof annotations.highlights[highlight.id] != 'undefined' ? annotations.highlights[highlight.id] : (typeof annotations.all[highlight.key] != 'undefined' ? annotations.all[highlight.key] : new Highlight());
            h.copyFrom(highlight);
            annotations.highlights[highlight.id] = h;
        }
        else
        {
            h = new Highlight();
            h.copyFrom(highlight);
            h.id = 0;
        }
        if (h != null)
            annotations.all[h.key] = h;
        return h;
    },
    addHighlights: function(highlights)
    {
        for (var key in highlights)
            annotations.addHighlight(highlights[key]);
    },
    addNote: function(note)
    {
        if (note.deleted)
            return null;
        var n = null;
        if (typeof note.id != 'undefined' && note.id > 0)
        {
            n = typeof annotations.notes[note.id] != 'undefined' ? annotations.notes[note.id] : (typeof annotations.all[note.key] != 'undefined' ? annotations.all[note.key] : new Note());
            n.copyFrom(note);
            annotations.notes[note.id] = n;
        }
        else
        {
            n = new Note();
            n.copyFrom(note);
            n.id = 0;
        }
        if (n != null)
            annotations.all[n.key] = n;
        return n;
    },
    addNotes: function(notes)
    {
        for (var key in notes)
            annotations.addNote(notes[key]);
    },
    count: function()
    {
        return annotations.all.count();
    },
    remove: function(annotation, wipe)
    {
        wipe = typeof wipe == 'boolean' ? wipe : false;
        if (wipe || annotation.deleted)
        {
            if (typeof publiwide.annotations.all[annotation.key] != 'undefined')
                delete publiwide.annotations.all[annotation.key];
            if (annotation.type == AnnotationType.Bookmark)
                publiwide.annotations.bookmarks.remove(annotation);
            else if (annotation.type == AnnotationType.Highlight)
                publiwide.annotations.highlights.remove(annotation);
            else if (annotation.type == AnnotationType.Note)
                publiwide.annotations.notes.remove(annotation);

            annotations.hide(annotation);
        }
        else
        {
            var annot = publiwide.annotations.all[annotation.key];
            annotation.deleted = true;
            annotation.modified = publiwide.utils.getDate(new Date()).format('yyyy-mm-dd HH:MM:ss', true);
            if(annot) {
                annot.deleted = true;
                annot.modified = publiwide.utils.getDate(new Date()).format('yyyy-mm-dd HH:MM:ss', true);
            }
        }
        publiwide.annotations.annotationRemoved.fire(annotation);
        publiwide.annotations.annotationRemoved.fire(annot);
    },
    toArray: function(sortProperty, ascending)
    {
        var result = [];
        var arr = annotations.all.toArray();
        for (var xi = 0, xl = arr.length; xi < xl; xi++)
            result.push(arr[xi]);

        for (var i = 0, l = result.length; i < l; i++)
        {
            for (var j = i + 1; j < l; j++)
            {
                var a = result[i];
                var b = result[j];
                if ((utils.getProperty(a, sortProperty) > utils.getProperty(b, sortProperty) && ascending) || (utils.getProperty(a, sortProperty) < utils.getProperty(b, sortProperty) && !ascending))
                {
                    result[j] = a;
                    result[i] = b;
                }
            }
        }

        return result;
    },

    isAnnotationOfPage: function(annotation){
        var annotationIdref = "";
        var data = annotation.data;
        if(typeof data != 'string'){
            data = data.toString();
        }
        if(data.indexOf("{") > -1 && data.indexOf("}") > -1){
            annotationIdref = data.split("{")[1].split("}")[0];
        }else if(data.indexOf("!") > -1){
            var firstPartData = data.split("!")[0];
            if(firstPartData.indexOf("[") > -1 && firstPartData.indexOf("]") > -1){
                annotationIdref = data.split("[")[1].split("]")[0];
            }
        }
	    var loadedSpineItems = ReadiumSDK.reader.getLoadedSpineItems();
	    for (i = 0; i < loadedSpineItems.length; i++){
		    var spineItem = loadedSpineItems[i];
		    if(spineItem.idref == annotationIdref){
			    return true;
		    }
	    }
        return false;
    },

    display: function(annotation, force)
    {
        force = typeof force == 'boolean' ? force : false;
        //if (annotation.visible && force)
           // annotations.hide(annotation);
        if (annotation.cType == 'epubcfi' /*&& !annotation.visible*/)
        {
            var cfi = new publiwide.CFI(annotation.data);
            if (cfi != null && cfi.parent != null && cfi.start != null && cfi.end != null)
            {
                if (publiwide.annotations.isAnnotationOfPage(annotation))
				{
                    if (annotation.type == publiwide.annotations.AnnotationType.Highlight)
					{
                        if (viewAnnot.isSVG())
                            annotations.displayHighlightSVG(annotation);
                        else
                            annotations.displayHighlight(annotation);
					}
                    else if (annotation.type == publiwide.annotations.AnnotationType.Note)
					{
                        if (viewAnnot.isSVG())
                            annotations.displayNoteSVG(annotation);
                        else
                            annotations.displayNote(annotation);
                    }
                    annotation.visible = true;
                }
            }
        }
    },
    displayHighlight: function(highlight)
    {
        if (typeof highlight.data == 'undefined')
            return;

        var doc = viewAnnot.getDocument();
        var first = true;
        var icon = doc.createElement('div');
        icon.href = 'javascript:;';
        icon.title = 'Click to edit the highlight';
        icon.id = 'aicon_' + highlight.key;
        icon.setAttribute('aid', highlight.key);
        icon.style.width = '16px';
        icon.style.height = '14px';
        icon.style.display = 'inline-block';
        //icon.style.background = "url('" + utils.baseURL() + "images/annotations/highlight_icons.png') no-repeat";
        icon.style.marginRight = '3px';
        icon.style.cursor = 'pointer';
        icon.highlight = highlight;
        icon.onclick = function()
        {
            annotations.highlightClick(this.highlight);
        };
        var color = '';
        switch (highlight.category)
        {
            default :
                color = 'rgba(232, 110, 114, 0.5)';
                icon.style.backgroundPosition = '0px';
                break;
            case 2:
                color = 'rgba(169, 102, 211, 0.5)';
                icon.style.backgroundPosition = '-17px';
                break;
            case 3:
                color = 'rgba(88, 195, 227, 0.5)';
                icon.style.backgroundPosition = '-34px';
                break;
            case 4:
                color = 'rgba(152, 181, 38, 0.5)';
                icon.style.backgroundPosition = '-51px';
                break;
            case 5:
                color = 'rgba(240, 216, 104, 0.5)';
                icon.style.backgroundPosition = '-68px';
                break;
        }

        var range = annotations.getRange(highlight.data);
        if (range == null)
            return;
        var selection = new publiwide.textSelection(range);

        for (var i = 0; i < selection.nodes.length; i++)
        {
            var node = selection.nodes[i];
            var nv = node.nodeValue;
            if (nv == null)
                continue;
            var beforeText = nv.substring(0, node.startOffset);
            var afterText = nv.substring(node.endOffset);
            var spanText = nv.substring(node.startOffset, node.endOffset);
            if (spanText.length > 0)
            {
                var span = doc.createElement('span');
                span.className = highlight.key;
                span.style.background = color;
                span.setAttribute('aid', highlight.key);

                var anchor = doc.createElement('a');
                anchor.title = 'Click to edit the highlight';
                anchor.style.color = 'inherit !important';
                anchor.style.textDecoration = 'none';
                anchor.setAttribute('aid', highlight.key);
                anchor.href = 'javascript:;';
                anchor.appendChild(doc.createTextNode(spanText));

                anchor.highlight = highlight;
                anchor.onclick = function()
                {
                    annotations.highlightClick(this.highlight);
                };
                span.appendChild(anchor);

                if (beforeText.length > 0)
                {
                    node.parentNode.insertBefore(document.createTextNode(beforeText), node);
                }
                node.parentNode.insertBefore(span, node);
                if (first && span.innerHTML.length > 0)
                {
                    span.id = highlight.key;
                    span.name = highlight.key;
                    node.parentNode.insertBefore(icon, span);
                    first = false;
                }

                if (afterText.length > 0)
                {
                    node.parentNode.insertBefore(document.createTextNode(afterText), node);
                }
                node.parentNode.removeChild(node);
            }
        }
    },
    displayHighlightSVG: function(highlight)
    {
        //highlight.rects = [];
        var color = '#e86e72';
        switch (highlight.category)
        {
            default:
                color = '#e86e72';
                break;
            case 2:
                color = '#a966d3';
                break;
            case 3:
                color = '#58c3e3';
                break;
            case 4:
                color = '#98b526';
                break;
            case 5:
                color = '#f0d868';
                break;
        }

        var range = annotations.getRange(highlight.data);
        if (range == null)
            return;
        var selection = new publiwide.textSelection(range);

        var boxes = [];
        for (var n = 0; n < selection.nodes.length; n++)
        {
            var node = selection.nodes[n];
            var text = node.nodeValue;
            var svgBox = utils.getSVGBBox(node, node.startOffset >= 0 ? node.startOffset : 0, node.endOffset >= 0 && node.endOffset < text.length ? node.endOffset : text.length - 1);

            if (utils.rectInArray(boxes, svgBox))
                continue;
            boxes.push(svgBox);
        }
        boxes = annotations._unifyLines(boxes);
        for (var i = 0; i < boxes.length; i++)
        {
            var box = boxes[i];
            var b = viewAnnot.getBody();
            var svg = b.getElementsByTagName('svg')[0];
            var a = utils.createOn(svg, 'a', { 'xlink:href': 'javascript:;', 'aid': highlight.key, 'id': 'aicon_' + highlight.key });
            a.appendChild(utils.createOn(svg, 'title', [], 'Click to edit the highlight'));
            a.highlight = highlight;
            a.onclick = function()
            {
                annotations.highlightClick(this.highlight);
            };

            a.style.cursor = 'pointer';

            var rect = viewAnnot.getDocument().createElementNS('http://www.w3.org/2000/svg', 'rect');
            rect.setAttributeNS(null, 'fill', color);
            rect.setAttributeNS(null, 'fill-opacity', '0.50');
            rect.setAttributeNS(null, 'x', box.left);
            rect.setAttributeNS(null, 'y', box.top);
            rect.setAttributeNS(null, 'width', box.width);
            rect.setAttributeNS(null, 'height', box.height);
            rect.setAttributeNS(null, 'aid', highlight.key);
            //highlight.rects.push(rect);
            a.appendChild(rect);

            if (viewAnnot.isSVG())
            {
                var parent = null;
                if (selection.nodes.length > 0 && selection.nodes[0].nodeName == 'tspan')
                {
                    parent = selection.nodes[0];
                    while (parent.nodeName != 'text' && parent.parentNode != null)
                        parent = parent.parentNode;
                }
                svg.appendChild(a);
            }
        }
    },
    displayNote: function(note)
    {
        var doc = viewAnnot.getDocument();
        var first = true;
        var icon = annotations._getNoteIcon(note);
        icon.note = note;
        icon.onclick = function()
        {
            annotations.noteClick(this.note);
        };
        var color = '';
        var linkTitle = 'Click to edit the note';
        if (note.isPublic && note.publicUserName != null && note.publicUserName.length > 0)
        {
            color = '#cccccc';
            linkTitle = 'Click to view the note created by ' + note.publicUserName;
        }
        else
        {
            switch (note.category)
            {
                default:
                    color = 'rgba(232, 110, 114, 0.5)';
                    icon.style.backgroundPosition = '0px';
                    break;
                case 2:
                    color = 'rgba(169, 102, 211, 0.5)';
                    if (note.isPublic)
                        icon.style.backgroundPosition = '-19px';
                    else
                        icon.style.backgroundPosition = '-13px';
                    break;
                case 3:
                    color = 'rgba(88, 195, 227, 0.5)';
                    if (note.isPublic)
                        icon.style.backgroundPosition = '-39px';
                    else
                        icon.style.backgroundPosition = '-26px';
                    break;
                case 4:
                    color = 'rgba(152, 181, 38, 0.5)';
                    if (note.isPublic)
                        icon.style.backgroundPosition = '-59px';
                    else
                        icon.style.backgroundPosition = '-39px';
                    break;
                case 5:
                    color = 'rgba(240, 216, 104, 0.5)';
                    if (note.isPublic)
                        icon.style.backgroundPosition = '-79px';
                    else
                        icon.style.backgroundPosition = '-52px';
                    break;
            }
        }

        var range = annotations.getRange(note.data);
        if (range == null)
            return;
        var selection = new publiwide.textSelection(range);
        for (var i = 0; i < selection.nodes.length; i++)
        {
            var node = selection.nodes[i];
            var nv = node.nodeValue;
            var beforeText = nv.substring(0, node.startOffset);
            var afterText = nv.substring(node.endOffset);
            var spanText = nv.substring(node.startOffset, node.endOffset);
            if (spanText.length > 0)
            {
                var span = doc.createElement('span');
                span.className = note.key;
                span.style.background = color;
                span.setAttribute('aid', note.key);

                var anchor = doc.createElement('a');
                anchor.title = linkTitle;
                anchor.style.color = 'inherit';
                anchor.style.textDecoration = 'none';
                anchor.setAttribute('aid', note.key);
                anchor.href = 'javascript:;';
                anchor.appendChild(doc.createTextNode(spanText));

                anchor.note = note;
                anchor.onclick = function()
                {
                    annotations.noteClick(this.note);
                };
                span.appendChild(anchor);

                if (beforeText.length > 0)
                    node.parentNode.insertBefore(doc.createTextNode(beforeText), node);
                node.parentNode.insertBefore(span, node);
                if (first && span.innerHTML.length > 0)
                {
                    span.id = note.key;
                    span.name = note.key;
                    node.parentNode.insertBefore(icon, span);
                    first = false;
                }

                if (afterText.length > 0)
                    node.parentNode.insertBefore(doc.createTextNode(afterText), node);
                node.parentNode.removeChild(node);
            }
        }
    },
    displayNoteSVG: function(note)
    {
        //note.rects = [];
        var color = '#e86e72';
        switch (note.category)
        {
            default:
                color = '#e86e72';
                break;
            case 2:
                color = '#a966d3';
                break;
            case 3:
                color = '#58c3e3';
                break;
            case 4:
                color = '#98b526';
                break;
            case 5:
                color = '#f0d868';
                break;
        }

        var range = annotations.getRange(note.data);
        if (range == null)
            return;
        var selection = new publiwide.textSelection(range);
        var first = true;

        var boxes = [];
        for (var n = 0; n < selection.nodes.length; n++)
        {
            var node = selection.nodes[n];
            var text = node.nodeValue;
            var svgBox = utils.getSVGBBox(node, node.startOffset >= 0 ? node.startOffset : 0, node.endOffset >= 0 && node.endOffset < text.length ? node.endOffset : text.length - 1);

            if (utils.rectInArray(boxes, svgBox))
                continue;
            boxes.push(svgBox);
        }
        boxes = annotations._unifyLines(boxes);
        var topLeftBox = utils.getTopLeft(boxes);
        for (var i = 0; i < boxes.length; i++)
        {
            var box = boxes[i];
            var b = viewAnnot.getBody();
            var svg = b.getElementsByTagName('svg')[0];
            var a = utils.createOn(svg, 'a', { 'xlink:href': 'javascript:;', 'aid': note.key, 'id': 'aicon_' + note.key });
            a.note = note;
            a.appendChild(utils.createOn(svg, 'title', [], 'Click to edit the note'));
            a.onclick = function()
            {
                annotations.noteClick(this.note);
            };

            a.style.cursor = 'pointer';

            var rect = viewAnnot.getDocument().createElementNS('http://www.w3.org/2000/svg', 'rect');
            rect.setAttributeNS(null, 'fill', color);
            rect.setAttributeNS(null, 'fill-opacity', '0.50');
            rect.setAttributeNS(null, 'x', box.left);
            rect.setAttributeNS(null, 'y', box.top);
            rect.setAttributeNS(null, 'width', box.width);
            rect.setAttributeNS(null, 'height', box.height);
            rect.setAttributeNS(null, 'aid', note.key);
            //note.rects.push(rect);
            a.appendChild(rect);

            if (first)
            {
                var icon = annotations._getNoteIconSVG(note);
                var iconWidth = parseFloat(icon.getAttribute('width'));
                icon.setAttributeNS(null, 'x', (String)(topLeftBox.left - iconWidth / 2));
                icon.setAttributeNS(null, 'y', (String)(topLeftBox.top - topLeftBox.height / 2));
                icon.setAttributeNS(null, 'height', topLeftBox.height);
                //note.icon = icon;
                a.appendChild(icon);
                first = false;
            }

            if (viewAnnot.isSVG())
            {
                var parent = null;
                if (selection.nodes.length > 0 && selection.nodes[0].nodeName == 'tspan')
                {
                    parent = selection.nodes[0];
                    while (parent.nodeName != 'text' && parent.parentNode != null)
                        parent = parent.parentNode;
                }
                svg.appendChild(a);
            }
        }
    },
    getHighlightFromUserSelection: function(sel, category)
    {
        if (sel == null)
            sel = annotations.getTextSelection();
        var result = null;
        if (sel != null && !sel.empty())
        {
            result = new Highlight(sel.getCFI());
            result.category = typeof category != 'undefined' ? category : 1;
            result.text = sel.text;
            sel.clear();
        }
        return result;
    },
    getNoteFromUserSelection: function(sel, category)
    {
        if (sel == null)
            sel = annotations.getTextSelection();
        var result = null;
        if (sel != null)
        {
            result = new Note(sel.getCFI());
            result.category = typeof category != 'undefined' ? category : 1;
            result.text = sel.text;
            sel.clear();
        }
        return result;
    },
	getBookmark: function()
	{
		var w = viewAnnot.getWindow();
		utils.getTopElement();
		var element = utils.topElement;
		var doc = viewAnnot.getDocument();
		if (element == null && typeof doc.caretRangeFromPoint == 'function')
		{
			var range1 = doc.caretRangeFromPoint(0, 0);
			if (range1 != null)
				element = range1.startContainer;
		}
		else if (element == null && typeof doc.caretPositionFromPoint == 'function')
		{
			var range2 = doc.caretPositionFromPoint(0, 0);
			element = range2.offsetNode;
		}
		var cfi = null;
		if (element != null)
		{
			var posX = utils.findPosX(element);
			var posY = utils.findPosY(element);
			var percentX = posX != 0 ? w.pageXOffset / posX : 0;
			var percentY = posY != 0 ? w.pageYOffset / posY : 0;
			cfi = new publiwide.EpubCFI(element, null, -1, null, -1, null, percentX, percentY);
		}
		if (cfi != null)
			return new Bookmark(cfi);
		return null;
	},
    getRange: function(data)
    {
        if (typeof data == 'object')
            data = data.toString();

        var cfi = new publiwide.CFI(data);

        if (cfi.parent != null)
        {
            var range = rangy.createRange();
			if (cfi.start != null)
			{
				range.startContainer = cfi.start.node;
				range.startOffset = cfi.start.index;
				var si = cfi.start.index;
				for (var s = 0; s < cfi.start.node.childNodes.length; s++)
				{
					var sNode = cfi.start.node.childNodes.item(s);
					var scc = utils.getCharCount(sNode);
					if (si - scc < 0)
					{
						range.startContainer = sNode;
						range.startOffset = si;
						break;
					}
					si -= scc;
				}
			}

			if (cfi.end != null)
			{
				range.endContainer = cfi.end.node;
				range.endOffset = cfi.end.index;
				var ei = cfi.end.index;
				for (var n = 0; n < cfi.end.node.childNodes.length; n++)
				{
					var node = cfi.end.node.childNodes.item(n);
					var cc = utils.getCharCount(node);
					if (ei - cc < 0)
					{
						range.endContainer = node;
						range.endOffset = ei;
						break;
					}
					ei -= cc;
				}
			}

            if (cfi.parent.node != null)
                range.commonAncestorContainer = cfi.parent.node;
            return range;
        }
        return null;
    },
    getTextSelection: function()
    {
        var selection = rangy.getSelection(viewAnnot.getWindow());
        if (selection == null || selection.rangeCount < 1)
            return null;

        var result = null;
        var range = selection.getRangeAt(0);
        if (range != null)
        {
            if (range.startContainer == range.endContainer && range.startOffset > range.endOffset)
            {
                var aux = range.startOffset;
                range.startOffset = range.endOffset;
                range.endOffset = aux;
            }
            result = new publiwide.textSelection(range);
            result.text = selection.toString();
        }
        return result;
    },
    highlightClick: function(highlight)
    {
        publiwide.androidAnnotations.highlightClick(highlight);
        return typeof highlight.onClick == "function" ? highlight.onClick() : false;

    },
    noteClick: function(note)
    {
        publiwide.androidAnnotations.noteClick(note);
        return typeof note.onClick == "function" ? note.onClick() : false;
    },
    hide: function(annotation)
    {
        if (viewAnnot.isSVG())
        {
            var anchors = viewAnnot.getDocument().getElementsByTagName('a');
            var i = 0;
            while (anchors.length > 0 && i < anchors.length)
            {
                var anchor = anchors[i];
                if (anchor.getAttribute('aid') != null && anchor.getAttribute('aid') == annotation.key && anchor.parentNode != null)
                    anchor.parentNode.removeChild(anchor);
                else
                    i++;
            }
        }
        var range = annotations.getRange(annotation.data);
        if (range != null)
        {
            annotations._removeAnnotationTags(range.commonAncestorContainer, annotation.key, 'aid');
            annotations._concatTextChildNodes(range.commonAncestorContainer);
        }

        var elem = viewAnnot.getDocument().getElementById('canvas' + annotation.key);
        if (elem != null)
            elem.parentNode.removeChild(elem);
        annotation.visible = false;
    },
    refresh: function(annotation)
    {
        annotations.hide(annotation);
        annotations.display(annotation);
    },
    refreshAll: function()
    {
        for (var key in annotations.all)
            annotations.refresh(annotations.all[key]);
    }
};

publiwide.annotations.Highlight = function(data)
{
    this._owner = true;
    this._stack = [];
    this.id = 0;
    this.created = utils.getDate(new Date()).format('yyyy-mm-dd HH:MM:ss', true);
    this.modified = utils.getDate(new Date()).format('yyyy-mm-dd HH:MM:ss', true);
    this.category = 1;
    this.cType = 'epubcfi';
    this.data = data;
    this.deleted = false;
    this.key = utils.generateUID();
    this.text = '';
    this.type = publiwide.annotations.AnnotationType.Highlight;
    this.visible = false;

    this.copyFrom = function(object)
    {
        if (typeof object == 'undefined')
            return;
        if (typeof object.id != 'undefined')
            this.id = object.id;
        if (typeof object.category != 'undefined')
            this.category = object.category;
        if (typeof object.data != 'undefined')
            this.data = object.data;
        if (typeof object.deleted != 'undefined')
            this.deleted = object.deleted;
        if (typeof object.text != 'undefined')
            this.text = object.text;
        if (typeof object.created != 'undefined')
            this.created = object.created;
        if (typeof object.modified != 'undefined')
            this.modified = object.modified;
        if (typeof object.visible != 'undefined')
            this.visible = object.visible;
    };
    this.getToc = function()
    {
        var data = this.data;
        if (typeof data != 'string')
            data = data.toString();
        if (data.startsWith('epubcfi(') && data.endsWith(')'))
            data = data.substr(8, data.length - 9);

        var i = data.indexOf('!');
        if (i > 0)
        {
            var url = data.substr(0, i);
            return toc.itemsCFI[url];
        }
        return null;
    };
    this.popState = function(empty)
    {
        empty = typeof empty != 'undefined' ? empty : false;
        if (this._stack.length > 0)
        {
            var state = this._stack.pop();
            if (!empty)
                this.copyFrom(state);
            return true;
        }
        return false;
    };
    this.pushState = function()
    {
        this._stack.push({
            id: this.id,
            modified: this.modified,
            category: this.category,
            data: this.data,
            deleted: this.deleted,
            text: this.text,
            visible: this.visible
        });
    };
    this.update = function()
    {
        this.modified = utils.getDate(new Date()).format('yyyy-mm-dd HH:MM:ss', true);;
    };
};
var Highlight = publiwide.annotations.Highlight;

publiwide.annotations.Note = function(data)
{
    this._owner = true;
    this._stack = [];
    this.id = 0;
    this.created = utils.getDate(new Date()).format('yyyy-mm-dd HH:MM:ss', true);
    this.modified = utils.getDate(new Date()).format('yyyy-mm-dd HH:MM:ss', true);
    this.category = 1;
    this.cType = 'epubcfi';
    this.data = data;
    this.deleted = false;
    this.isPublic = false;
    this.key = utils.generateUID();
    this.publicUserName = '';
    this.text = '';
    this.type = publiwide.annotations.AnnotationType.Note;
    this.userText = '';
    this.visible = false;

    this.copyFrom = function(object)
    {
        if (typeof object == 'undefined')
            return;

        if (typeof object.id != 'undefined')
            this.id = object.id;
        if (typeof object.category != 'undefined')
            this.category = object.category;
        if (typeof object.data != 'undefined')
            this.data = object.data;
        if (typeof object.deleted != 'undefined')
            this.deleted = object.deleted;
        if (typeof object.isPublic != 'undefined')
            this.isPublic = object.isPublic;
        if (typeof object.text != 'undefined')
            this.text = object.text;
        if (typeof object.userText != 'undefined')
            this.userText = object.userText;
        if (typeof object.created != 'undefined')
            this.created = object.created;
        if (typeof object.modified != 'undefined')
            this.modified = object.modified;
    };
    this.getToc = function()
    {
        var data = this.data;
        if (typeof data != 'string')
            data = data.toString();
        if (data.startsWith('epubcfi(') && data.endsWith(')'))
            data = data.substr(8, data.length - 9);

        var i = data.indexOf('!');
        if (i > 0)
        {
            var url = data.substr(0, i);
            return toc.itemsCFI[url];
        }
        return null;
    };
    this.popState = function(empty)
    {
        empty = typeof empty != 'undefined' ? empty : false;
        if (this._stack.length > 0)
        {
            var state = this._stack.pop();
            if (!empty)
                this.copyFrom(state);
            return true;
        }
        return false;
    };
    this.pushState = function()
    {
        this._stack.push({
            id: this.id,
            modified: this.modified,
            category: this.category,
            data: this.data,
            deleted: this.deleted,
            text: this.text,
            visible: this.visible,
            userText: this.userText,
            isPublic: this.isPublic
        });
    };
    this.update = function()
    {
        this.modified = utils.getDate(new Date()).format('yyyy-mm-dd HH:MM:ss', true);
    };
};
var Note = publiwide.annotations.Note;

publiwide.annotations.Bookmark = function(data)
{
	this._stack = [];
	this.id = 0;
	this.created = utils.getDate(new Date()).format('yyyy-mm-dd HH:MM:ss', true);
	this.modified = utils.getDate(new Date()).format('yyyy-mm-dd HH:MM:ss', true);
	this.category = 1;
	this.cType = 'epubcfi';
	this.data = data;
	this.deleted = false;
	this.key = utils.generateUID();
	this.text = '';
	this.type = publiwide.annotations.AnnotationType.Bookmark;
	this.visible = false;

	this.copyFrom = function(object)
	{
		if (typeof object == 'undefined')
			return;

		if (typeof object.id != 'undefined')
			this.id = object.id;
		if (typeof object.key != 'undefined')
			this.key = object.key;
		if (typeof object.category != 'undefined')
			this.category = object.category;
		if (typeof object.data != 'undefined')
			this.data = object.data;
		if (typeof object.deleted != 'undefined')
			this.deleted = object.deleted;
		if (typeof object.text != 'undefined')
			this.text = object.text;
		if (typeof object.created != 'undefined')
			this.created = object.created;
		if (typeof object.modified != 'undefined')
			this.modified = object.modified;
		if (typeof object.visible != 'undefined')
			this.visible = object.visible;
	};
	this.getToc = function()
	{
		var data = this.data;
		if (typeof data != 'string')
			data = data.toString();
		if (data.startsWith('epubcfi(') && data.endsWith(')'))
			data = data.substr(8, data.length - 9);

		var i = data.indexOf('!');
		if (i > 0)
		{
			var url = data.substr(0, i);
			return toc.itemsCFI[url];
		}
		return null;
	};
	this.popState = function(empty)
	{
		empty = typeof empty != 'undefined' ? empty : false;
		if (this._stack.length > 0)
		{
			var state = this._stack.pop();
			if (!empty)
				this.copyFrom(state);
			return true;
		}
		return false;
	};
	this.pushState = function()
	{
		this._stack.push({
			id: this.id,
			modified: this.modified,
			category: this.category,
			data: this.data,
			deleted: this.deleted,
			text: this.text,
			visible: this.visible
		});
	};
	this.update = function()
	{
		this.modified = utils.getDate(new Date()).format('yyyy-mm-dd HH:MM:ss', true);;
	};
};
var Bookmark = publiwide.annotations.Bookmark;

publiwide.annotations.AnnotationType =
{
    Highlight : 0,
    Note : 1,
    PublicNote : 2,
	Bookmark: 3
};
var AnnotationType = publiwide.annotations.AnnotationType;

publiwide.textSelection = function(range)
{
    this.endElement = null;
    this.endOffset = null;
    this.nodes = [];
    this.parentElement = null;
    this.range = null;
    this.startElement = null;
    this.startOffset = null;
    this.text = '';

    this._applyRange = function(range)
    {
        this.range = range;

        this.startElement = range.startContainer;
        while (this.startElement != null && this.startElement.nodeType != 1)
            this.startElement = this.startElement.parentNode;
        this.startOffset = this._getIndex(this.startElement, range.startContainer, range.startOffset);

        this.endElement = range.endContainer;
        while (this.endElement != null && this.endElement.nodeType != 1)
            this.endElement = this.endElement.parentNode;
        this.endOffset = this._getIndex(this.endElement, range.endContainer, range.endOffset);
        this.parentElement = utils.getCommonAncestor(this.startElement, this.endElement);

        this.nodes = this.getNodes();
    };
    this._containsChild = function(parentNode, childNode)
    {
        var i = 0;
        while (i < parentNode.childNodes.length)
        {
            var child = parentNode.childNodes[i];
            if (child == childNode || this._containsChild(child, childNode))
                return true;
            i++;
        }
        return false;
    };
    this._getNextNode = function(node)
    {
        var result = null;
        if (node.hasChildNodes())
            result = node.childNodes[0];
        else
            result = node.nextSibling;
        while (result == null)
        {
            node = node.parentNode;
            if (node != null)
                result = node.nextSibling;
            else
                break;
        }
        if (result != null && result.nodeType != 3)
            return this._getNextNode(result);
        return result;
    };
    this._getIndex = function(parentNode, childNode, childIndex)
    {
        var result = 0;
        var i = 0;
        while (i < parentNode.childNodes.length)
        {
            var child = parentNode.childNodes[i];
            if (child == childNode)
                return result + childIndex;
            else
            {
                if (this._containsChild(child, childNode))
                    return result + this._getIndex(child, childNode, childIndex);
                else
                    result += utils.getCharCount(child);
            }
            i++;
        }
        return result;
    };
    this._rangeIntersectsNode = function(range, node)
    {
        if (node.nodeValue != null && node.nodeValue.trim().length == 0)
            return false;
        if (typeof window.Range != 'undefined' && Range.prototype.intersectsNode)
            return this._rangeIntersectsNode2(range, node);
        else
            return this._rangeIntersectsNode3(range, node);
    };
    this._rangeIntersectsNode2 = function(range, node)
    {
        if (range != null)
            return range.intersectsNode(node);
        return true;
    };
    this._rangeIntersectsNode3 = function(range, node)
    {
        var nodeRange = node.ownerDocument.createRange();
        try
        {
            nodeRange.selectNode(node);
        }
        catch (e)
        {
            nodeRange.selectNodeContents(node);
        }

        return range.compareBoundaryPoints(Range.END_TO_START, nodeRange) == -1 && range.compareBoundaryPoints(Range.START_TO_END, nodeRange) == 1;
    };


    this.clear = function()
    {
        var w = viewAnnot.getWindow();
        if (w.getSelection)
        {
            if (w.getSelection().empty)
                w.getSelection().empty();
            else if (w.getSelection().removeAllRanges)
                w.getSelection().removeAllRanges();
        }
        else if (getDocument().selection)
            getDocument().selection.empty();
    };
    this.empty = function()
    {
        return (this.startElement == this.endElement && this.startOffset == this.endOffset) || this.text.length == 0;
    };
    this.getCFI = function()
    {
        if ((this.parentElement == null || this.startElement == null || this.endElement == null) && this.range != null)
            this._applyRange(this.range);
        //return new publiwide.EpubCFI(this.parentElement, this.startElement, this.startOffset, this.endElement, this.endOffset);
        return new publiwide.EpubCFI(this.range.commonAncestorContainer, this.range.startContainer, this.range.startOffset, this.range.endContainer, this.range.endOffset);
    };
    this.getNodes = function()
    {
        if (this.nodes.length > 0)
            return this.nodes;
        var currentNode = this.range.startContainer;
        currentNode.startOffset = this.range.startOffset;

        while (currentNode != null && currentNode != this.range.endContainer)
        {
            currentNode.endOffset =  utils.getCharCount(currentNode);
            if (currentNode.endOffset > currentNode.startOffset && utils.getCharCountNWS(currentNode) > 0)
                this.nodes.push(currentNode);
            currentNode = this._getNextNode(currentNode);
            if (currentNode != null)
                currentNode.startOffset = 0;
        }
        if (currentNode != null)
        {
            currentNode.endOffset = this.range.endOffset;
            if (currentNode.endOffset > currentNode.startOffset && utils.getCharCountNWS(currentNode) > 0)
                this.nodes.push(currentNode);
        }

        return this.nodes;
    };
    this.toString = function()
    {
        return this.getCFI();
    };

    if (range != null)
        this._applyRange(range);
};

var annotations = publiwide.annotations;