/*
 Copyright (c) 2014 Publiwide
 http://www.publiwide.com

 Created: 9/10/14 12:17 PM
 Authors: Alexandru Popa, alexandru.popa@publiwide.com
 */


var publiwide = publiwide || {};
publiwide.currentCFI = 'epubcfi(/6/4/!0:0)';

publiwide.lastReadBookmarkId = "0";

publiwide.CFI = function(data)
{
	this.value = data;
	this.toc = null;
	this.fileCFI = '';
	this.documentCFI = '';
	this.parent = null;
	this.start = null;
	this.end = null;
	this.spatial = null;

	this._getLocation = function(data, parent)
	{
		if (typeof parent == 'undefined')
			parent = viewAnnot.getBody().parentNode;
		var ps = data.split('/');
		var result =
		{
			node: parent,
			index: 0
		};
		for (var i = 0; i < ps.length; i++)
		{
			var p = ps[i];
			if (p.length > 0)
			{
				var c = p.indexOf(':');
				if (c > 0)
				{
					result.index = parseInt(p.substr(c + 1));
					p = p.substr(0, c);

					var index = (parseInt(p) - 1) / 2;
					var cn = 0;
					while (index >= 0 && cn < result.node.childNodes.length)
					{
						if (result.node.childNodes[cn++].nodeType == 3)
							index--;
					}
					cn--;
					if (result.node.childNodes.length > 0 && cn >= 0 && cn < result.node.childNodes.length)
						result.node = result.node.childNodes[cn];
					else
						return null;
				}
				else
				{
					result.index = 0;
					var index = parseInt(p) / 2 - 1;
					if (result.node.children.length > 0 && index >= 0 && index < result.node.children.length)
						result.node = result.node.children[index];
					else
						return null;
				}
			}
		}
		return result;
	};

	//todo: error prone function, more validation please
	this.init = function(data)
	{
		if (typeof data == 'object')
		{
			if (typeof data.toString == 'function')
				data = data.toString();
			else if (typeof data.fullValue == 'string')
				data = data.fullValue;
			else if (typeof data.value == 'string')
				data = data.value;
		}

		var tcfi = data;

		if (data.startsWith('epubcfi(') && data.endsWith(')'))
			data = data.substr(8, data.length - 9);

		var i = tcfi.indexOf('!');
		if (i > 0)
		{
			this.fileCFI = tcfi.substr(0, i);
			if (!this.fileCFI.endsWith(')'))
				this.fileCFI += ')';
			//var k = this.fileCFI.endsWith(')') ? this.fileCFI : this.fileCFI + ')';
			//this.toc = toc.itemsCFI[this.fileCFI] || toc.itemsCFI[this.fileCFI.substr(8, this.fileCFI.length - 9)];
		}
		i = data.indexOf('!');
		if (i > 0)
			this.documentCFI = data.substr(i + 1);
		else
			this.documentCFI = data;

		var idx = this.documentCFI.indexOf('@');
		if (idx > 0)
		{
			var parts = this.documentCFI.substr(idx + 1).split(':');
			if (parts.length > 1)
			{
				this.spatial =
				{
					X: parseFloat(parts[0]),
					Y: parseFloat(parts[1])
				};
				this.documentCFI = this.documentCFI.substr(0, idx);
			}
		}

		//todo: Location Correction (http://www.idpf.org/epub/linking/cfi/epub-cfi.html#sec-target-correction)
		var parts = this.documentCFI.split(',');
		if (parts.length == 1)
		{
			this.parent = this._getLocation(this.documentCFI);
			this.start = this.parent;
		}
		else if (parts.length == 3)
		{
			this.parent = this._getLocation(parts[0]);
			if (this.parent != null && this.parent.node != null)
			{
				this.start = this._getLocation(parts[1], this.parent.node);
				this.end = this._getLocation(parts[2], this.parent.node);
			}
		}
	};
	//todo: test
	this.equals = function(other)
	{
		if (typeof other == 'string')
			other = new publiwide.CFI(other);
		else if (typeof other.documentCFI == 'undefined' && typeof other.value == 'string')
			other = new publiwide.CFI(other.value);
		return this.documentCFI.replace("epubcfi(", "") == other.documentCFI.replace("epubcfi(", "") && this.parent == other.parent && this.start == other.start && this.end == other.end;
	};
	this.sameDocument = function(other)
	{
		if (typeof other == 'string')
			other = new publiwide.CFI(other);
		else if (typeof other.documentCFI == 'undefined' && typeof other.value == 'string')
			other = new publiwide.CFI(other.value);
		return this.documentCFI.replace("epubcfi(", "") == other.documentCFI.replace("epubcfi(", "");
	};

	this.init(data);

	/*if (typeof data == "object")
	 data = data.toString();
	 if (data.startsWith("epubcfi(") && data.endsWith(")"))
	 data = data.substr(8, data.length - 9);
	 this.value = data;
	 var i = this.value.indexOf("!");
	 if (i > 0)
	 {
	 var tcfi = this.value.substr(0, i);
	 if (tcfi.startsWith('{') && tcfi.endsWith('}'))
	 this.idref = tcfi.substr(1, tcfi.length - 2);
	 this.value = this.value.substr(i + 1);
	 }*/
};

/**
 * Represents a CFI.
 */
publiwide.EpubCFI = function(parentNode, startNode, startIndex, endNode, endIndex, temporal, spatialX, spatialY)
{
	this.excluded = function(node)
	{
		return typeof node != typeof undefined && typeof node.hasAttribute == 'function' && node.hasAttribute('data-cfi-exclude') && node.getAttribute('data-cfi-exclude').toLowerCase() != 'false';
	};
	this.getIndex = function(node)
	{
		if (typeof node != typeof undefined && typeof node.parentNode != typeof undefined)
		{
			//return Array.prototype.indexOf.call(node.parentNode.children, node) + 1;
			for (var i = 0; i < node.parentNode.children.length; i++)
			{
				var child = node.parentNode.children[i];
				if (!this.excluded(child) && child == node)
					return i + 1;
			}
		}
		console.log('invalid data?');
		return 0;
	};
	this.findIndex = function(node)
	{
		var result = 0;
		if (node != null && node.parentNode != null)
		{
			for (var i = 0; i < node.parentNode.childNodes.length; i++)
			{
				var child = node.parentNode.childNodes[i];
				if (child == node)
					break;
				if (child.nodeType == node.nodeType)
					result++;
			}
		}
		return result;
	};
	this.getCFINode = function(node)
	{
		if (node == null || node.parentNode == null || node.parentNode.parentNode == null)
			return '';
		var result = this.getCFINode(node.parentNode);
		if (node.nodeType == 1 && !this.excluded(node))
			result += '/' + (this.getIndex(node) * 2) + (node.hasAttribute('id') ? '[' + node.getAttribute('id') + ']' : '');
		return result;
	};
	this.getCFI = function(node, index)
	{
		var result = this.getCFINode(node);
		if (index >= 0)
			result += '/' + (this.findIndex(node) * 2 + 1) + ':' + index;
		return result;
	};
	this.toString = function(full)
	{
		return typeof full == 'boolean' ? (full ? this.fullValue : this.value) : this.fullValue;
	};

	var parentCFI = '';
	var startCFI = '';
	var endCFI = '';
	if (typeof parentNode != 'undefined' && parentNode != null && (typeof startNode == 'undefined' || startNode == null) && (typeof endNode == 'undefined' || endNode == null))
	{
		parentCFI = this.getCFI(parentNode, -1);
		this.value = 'epubcfi(' + parentCFI;
	}
	else
	{
		if (parentNode == null || typeof parentNode == 'undefined')
			parentNode = utils.getCommonAncestor(startNode, endNode);
		if (parentNode != null && (parentNode == startNode || parentNode == endNode))
			parentNode = parentNode.parentNode;
		parentCFI = this.getCFI(parentNode, -1);
		startCFI = this.getCFI(startNode, startIndex);
		if (startCFI.startsWith(parentCFI))
			startCFI = startCFI.substr(parentCFI.length);
		endCFI = this.getCFI(endNode, endIndex);
		if (endCFI.startsWith(parentCFI))
			endCFI = endCFI.substr(parentCFI.length);

		if (endCFI.length > 0)
			this.value = 'epubcfi(' + parentCFI + (startCFI.length > 0 ? ',' + startCFI + (endCFI.length > 0 ? ',' + endCFI : '') : '');
		else
			this.value = 'epubcfi(' + startCFI;
	}

	var rs = undefined;
	if (typeof ReadiumSDK != 'undefined' && typeof ReadiumSDK.reader != 'undefined' && typeof ReadiumSDK.reader.getCurrentSelectionCfi == 'function')
		rs = ReadiumSDK.reader.getCurrentSelectionCfi();
	this.fullValue = (typeof rs == 'object' && typeof rs.idref == 'string' && rs.idref.length > 0) ? ('epubcfi({' + rs.idref + '}!' + parentCFI + ',' + startCFI + ',' + endCFI + ')') : this.value;
	if (typeof temporal == 'number')
	{
		this.value += '~' + temporal.toFixed(4);
		this.fullValue += '~' + temporal.toFixed(4);
	}
	if (typeof spatialX == 'number' && typeof spatialY == 'number')
	{
		this.value += '@' + spatialX.toFixed(4) + ':' + spatialY.toFixed(4);
		this.fullValue += '@' + spatialX.toFixed(4) + ':' + spatialY.toFixed(4);
	}
	this.value += ')';
	this.fullValue += ')';

	this.startPosition = utils.findPosY(startNode);
	this.endPosition = utils.findPosY(endNode);
};

publiwide.sync =
{
    attempts: 0,
    interval: 5000,
    key: null,
    running: false,
    deviceId: "",
    bookId: "",

    _completed: function () {
        publiwide.sync.running = false;
        setTimeout(function () {
            publiwide.sync.start(true);
        }, publiwide.sync.interval);
    },
    _getData: function () {
        var result = {};
        var hc = 0;
        var hs = {};
        var nhs = [];
        var nc = 0;
        var ns = {};
        var nns = [];
        var annots = publiwide.annotations.all.toArray();
        //var lastReadBookmark = publiwide.annotations.getBookmark();
        //lastReadBookmark.bookmarkType = 1;
        //lastReadBookmark.id = publiwide.lastReadBookmarkId;
        for (var i = 0; i < annots.length; i++) {
            var annot = annots[i];
            if (annot.type == AnnotationType.Highlight) {
                if (annot.id > 0) {
                    hs[annot.id] = annot;
                    hc++;
                }
                else
                    nhs.push(annot);
            }
            else if (annot.type == AnnotationType.Note) {
                if (annot.id > 0) {
                    ns[annot.id] = annot;
                    nc++;
                }
                else
                    nns.push(annot);
            }
        }
        if (hc > 0 || nc > 0 || nhs.length > 0 || nns.length > 0)
            result.annotations = {};
        if (hc > 0)
            result.annotations.highlights = hs;
        if (nc > 0)
            result.annotations.notes = ns;
        if (nhs.length > 0)
            result.annotations.newHighlights = nhs;
        if (nns.length > 0)
            result.annotations.newNotes = nns;
        /*if(typeof result.annotations == 'undefined'){
            result.annotations = {};
        }
        result.annotations.lastReadBookmark = lastReadBookmark;*/
        return result;
    },
    _init: function (deviceId, bookId) {
        publiwide.sync.deviceId = deviceId;
        publiwide.sync.bookId = bookId;
        var result = utils.generateUID();
        publiwide.web.services.call('sync', 'initSync', {
            'key': result,
            'data': JSON.stringify(publiwide.sync._getData())
        }, function (r) {
            if (r.successful()) {
                publiwide.sync.running = false;
                publiwide.sync.key = result;
                publiwide.sync.start();
            }
            else
                publiwide.sync._completed();
        });
    },

    start: function (full) {
        if (publiwide.sync.running && publiwide.sync.attempts < 5) {
            publiwide.sync.attempts++;
            return;
        }
        publiwide.sync.running = true;
        publiwide.sync.attempts = 0;
        full = typeof full == 'boolean' ? full : false;
        if (publiwide.sync.key == null)
            publiwide.sync.key = publiwide.sync._init();
        else {
            var dataToSend = "{\"" + publiwide.sync.bookId + "\":" + JSON.stringify(publiwide.sync._getData()) + "}";
            var data = {'deviceId': publiwide.sync.deviceId, 'key': publiwide.sync.key, 'data': dataToSend};
            if (full)
                data.type = 'full';
            publiwide.web.services.call('sync', 'sync', data, function (r) {
                try {
                    if (r.status == publiwide.web.WebServiceStatus.OK && typeof r.result != 'undefined') {
                        var toAdd = [];
                        var toRemove = [];
                        var resultForBook = r.result[publiwide.sync.bookId];

                        if (typeof resultForBook.annotations != 'undefined') {
                            //todo: bookmarks
                            if (typeof resultForBook.annotations.highlights != 'undefined') {
                                for (var h in resultForBook.annotations.highlights) {
                                    h = resultForBook.annotations.highlights[h];
                                    if (h.deleted) {
                                        h = publiwide.annotations.highlights[h.id];
                                        if (h != null)
                                            toRemove.push(h);
                                    }
                                    else {
                                        h = publiwide.annotations.addHighlight(h);
                                        if (h != null) {
                                            if (h.deleted) {
                                                toRemove.push(h);
                                            }
                                            else {
                                                toAdd.push(h);
                                                publiwide.androidAnnotations.syncHighlight(h);
                                            }
                                        }
                                    }
                                }
                            }
                            if (typeof resultForBook.annotations.notes != 'undefined') {
                                for (var n in resultForBook.annotations.notes) {
                                    n = publiwide.annotations.addNote(resultForBook.annotations.notes[n]);
                                    if (n != null) {
                                        if (n.deleted) {
                                            toRemove.push(n);
                                        }
                                        else {
                                            toAdd.push(n);
                                            publiwide.androidAnnotations.syncNote(n);
                                        }
                                    }
                                }
                            }
                            /*if(typeof resultForBook.annotations.lastReadBookmark != 'undefined'){
                                if(typeof resultForBook.annotations.lastReadBookmark.id != 'undefined') {
                                    publiwide.lastReadBookmarkId = resultForBook.annotations.lastReadBookmark.id;
                                }
                            }*/

                            //for (var a = 0; a < toAdd.length; a++)
                            //publiwide.annotations.display(toAdd[a]);
                            for (var tr = 0; tr < toRemove.length; tr++) {
                                publiwide.annotations.remove(toRemove[tr]);
                                publiwide.androidAnnotations.removeAnnotationSync(toRemove[tr]);
                            }
                        }
                    }
                }
                catch (ex) {
                    //if (publiwide.debug)
                    //	alert(ex);
                }
                finally {
                    publiwide.sync._completed();
                }
            });
        }
    }
};