/*
 Copyright (c) 2014 Publiwide
 http://www.publiwide.com

 Created: 9/9/14 12:15 PM
 Authors: Alexandru Popa, alexandru.popa@publiwide.com
 */

var publiwide = publiwide || {};
publiwide.utils =
{
    baseURL: function()
    {
        return location.href.substr(0, location.href.lastIndexOf("/") + 1);
    },
    createOn: function(root, name, attrs, text)
    {
        var doc = root.ownerDocument;
        var svg = root.ownerSVGElement || root;
        var svgNS = svg.getAttribute('xmlns');
        var el = doc.createElementNS(svgNS, name);
        for (var attr in attrs)
        {
            if (!attrs.hasOwnProperty(attr))
                continue;
            var parts = attr.split(':');
            if (parts[1])
                el.setAttributeNS(svg.getAttribute('xmlns:' + parts[0]), parts[1], attrs[attr]);
            else
                el.setAttributeNS(null, attr, attrs[attr]);
        }
        if (text)
            el.appendChild(document.createTextNode(text));
        return root.appendChild(el);
    },
    generateUID: function(separator)
    {
        separator = separator || "-";
        function S4()
        {
            return (((1 + Math.random()) * 0x10000) | 0).toString(16).substring(1);
        }
        return S4() + S4() + separator + S4() + separator + S4() + separator + S4() + separator + S4() + S4() + S4();
    },
    getCharCount: function(node)
    {
        if (node == null)
            return 0;
        if (node.nodeType == 3)
            return node.nodeValue.length;
        var result = 0;
        for (var t = 0; t < node.childNodes.length; t++)
        {
            var child = node.childNodes[t];
            if (child.nodeType == 3)
                result += child.nodeValue.length;
            else
                result += utils.getCharCount(child);
        }
        return result;
    },
    getCharCountNWS: function(node)
    {
        if (node == null)
            return 0;
        if (node.nodeType == 3)
            return node.nodeValue.trim().length;
        var result = 0;
        for (var t = 0; t < node.childNodes.length; t++)
        {
            var child = node.childNodes[t];
            if (child.nodeType == 3)
                result += child.nodeValue.trim().length;
            else
                result += utils.getCharCount(child);
        }
        return result;
    },
    getCommonAncestor: function(a, b)
    {
        var parents1 = utils.parents(a);
        var parents2 = utils.parents(b);

        if (parents1.length == 0 || parents2.length == 0)
            throw "Oups! Something wrong happened.";
        if (parents1[0] != parents2[0])
            return null;

        for (var i = 0; i < Math.min(parents1.length, parents2.length); i++)
            if (parents1[i] != parents2[i])
                return parents1[i - 1];

        var arr = parents1.length <= parents2.length ? parents1 : parents2;
        var idx = arr.length;
        while (--idx >= 0)
        {
            if (arr[idx] != a && arr[idx] != b)
                return arr[idx];
        }
        return arr[arr.length - 1];
    },
    getDate: function(date)
    {
        if (typeof date == 'string')
        {
            var year = parseInt(date.substring(0, 4).trimLeft(' 0'));
            var month = parseInt(date.substring(5, 7).trimLeft(' 0'));
            var day = parseInt(date.substring(8, 10).trimLeft(' 0'));
            var hours = parseInt(date.substring(11, 13).trimLeft(' 0'));
            var minutes = parseInt(date.substring(14, 16).trimLeft(' 0'));
            var seconds = parseInt(date.substring(17, 19).trimLeft(' 0'));
            return new Date(isNaN(year) ? 0 : year, isNaN(month - 1) ? 0 : month - 1, isNaN(day) ? 0 : day, isNaN(hours) ? 0 : hours, isNaN(minutes) ? 0 : minutes, isNaN(seconds) ? 0 : seconds, 0);
        }
        else if (typeof date == 'object')
            return date;
        else
            throw new SyntaxError('invalid date');
    },
    getOffset: function(element)
    {
        var box = { top: 0, left: 0 };
        var document = viewAnnot.getDocument();
        var window = viewAnnot.getWindow();

        if (typeof element.getBoundingClientRect !== 'undefined')
            box = element.getBoundingClientRect();
        return {
            top: box.top  + (window.pageYOffset || document.scrollTop)  - (document.clientTop || 0),
            left: box.left + (window.pageXOffset || document.scrollLeft) - (document.clientLeft || 0)
        };
    },
    getSVGBBox: function(element, start, end)
    {
        while (element.nodeType != 1 && element.parentNode != null)
            element = element.parentNode;
        var offset = $(element).offset();
        var bbox =
        {
            left: offset.left,
            top: offset.top,
            width: $(element).width(),
            height: $(element).height(),
            right: offset.left + $(element).width(),
            bottom: offset.top + $(element).height()
        };
        var rect = {};
        if (bbox.left > 0 || bbox.top > 0 || bbox.width > 0 || bbox.height > 0)
        {
            var b = viewAnnot.getBody();
            var svg = b.getElementsByTagName('svg')[0];
            var pt = svg.createSVGPoint();
            var screenToSVG = svg.getScreenCTM().inverse();

            pt.x = bbox.left;
            pt.y = bbox.top;
            var xy = pt.matrixTransform(screenToSVG);
            rect.left = xy.x;
            rect.top = xy.y;
            pt.x = bbox.right;
            pt.y = bbox.bottom;
            var xy2 = pt.matrixTransform(screenToSVG);
            rect.width = xy2.x - xy.x;
            rect.height = xy2.y - xy.y;
        }
        else
        {
            rect.left = element.offsetLeft;
            rect.top = element.offsetTop;
            rect.width = element.offsetWidth;
            rect.height = element.offsetHeight;
        }

        var elementWithX = element;
        var k = 3;
        while (elementWithX.getAttribute('x') == null && elementWithX.parentNode != null && k-- > 0)
            elementWithX = elementWithX.parentNode;
        var elementWithY = element;
        k = 3;
        while (elementWithY.getAttribute('y') == null && elementWithY.parentNode != null && k-- > 0)
            elementWithY = elementWithY.parentNode;

        if (elementWithX.getAttribute('x') != null)
            rect.left = parseFloat(elementWithX.getAttribute('x'));
        if (elementWithY.getAttribute('y') != null && element.offsetHeight > 0)
            rect.top = parseFloat(elementWithY.getAttribute('y')) - element.offsetHeight + element.offsetHeight * 0.1;

        if (start != null && end != null && element.attributes != null && element.attributes['x'] != null)
        {
            var Xs = element.attributes['x'].value.split(' ');
            var c = 0;
            Xs[0] = parseFloat(Xs[0]);
            for (var i = 1; i < Xs.length; i++)
            {
                Xs[i] = parseFloat(Xs[i]);
                c += Xs[i] - Xs[i - 1];
            }
            var med = c / (Xs.length - 1);
            if (Xs.length > end)
            {
                var s = Xs[0];
                rect.left += Xs[start] - s;
                if (end < Xs.length - 1)
                    rect.width = Xs[end] - Xs[start];
                else if (start > 0)
                    rect.width = Xs[end] + med - Xs[start];
            }
        }

        //rect.left += 2;
        //rect.top += 2;
        rect.height += 2;
        //rect.width += 2;
        return rect;
    },
    getTopLeft: function(array)
    {
        var result = null;
        for (var i = 0; i < array.length; i++)
        {
            var rect = array[i];
            if (result == null)
                result = rect;
            else
            {
                if (rect.top < result.top)
                    result = rect;
                else if (Math.abs(rect.top - result.top) <= 2 && rect.left < result.left)
                    result = rect;
            }
        }
        return result;
    },
    findPosX: function(element)
    {
        if (element == null)
            return 0;
        var curLeft = 0;
        if (element.offsetParent)
        {
            while (element.offsetParent)
            {
                curLeft += element.offsetLeft;
                element = element.offsetParent;
            }
        }
        else if (element.x)
            curLeft += element.x;
        return curLeft;
    },
    findPosY: function(element)
    {
        if (element == null)
            return 0;
        var curTop = 0;
        try
        {
            if (element.offsetParent)
            {
                while (element.offsetParent)
                {
                    curTop += element.offsetTop;
                    element = element.offsetParent;
                }
            }
            else if (element.y)
                curTop += element.y;
        }
        catch (Exception)
        {

        }
        return curTop;
    },
    offset: function(element)
    {
        var l = 0;
        var t = 0;
        if (element.offsetParent)
        {
            do
            {
                l += element.offsetLeft;
                t += element.offsetTop;

            } while (element = element.offsetParent);
        }

        return {left: l, top: t};
    },
    parents: function(node)
    {
        var result = [];
        for (; node; node = node.parentNode)
            result.unshift(node);
        return result;
    },
    rectInArray: function(array, rect)
    {
        for (var i = 0; i < array.length; i++)
        {
            var r = array[i];
            if (r.left == rect.left && r.top == rect.top && r.width == rect.width && r.height == rect.height)
                return true;
        }
        return false;
    },

    topElement: null,
    topElementPos: -999999,
    getTopElement: function(element)
    {
        if (typeof element == "undefined")
        {
            element = viewAnnot.getBody();
            utils.topElement = element;
            utils.topElementPos = -999999;
        }

        var y0 = viewAnnot.getWindow().pageYOffset;
        var dist = Math.abs(utils.findPosY(element) - y0);
        if (element != viewAnnot.getBody() && dist >= 0 && dist < Math.abs(utils.topElementPos))
        {
            utils.topElement = element;
            utils.topElementPos = dist;
        }

            //if (Math.abs(utils.topElementPos) >= 0)
            for (var i = 0; i < element.childNodes.length; i++) {
                utils.getTopElement(element.childNodes[i]);
            }

        return utils.topElement;
    }
};
var utils = publiwide.utils;

publiwide.utils.dateMasks =
{
    "default":      "ddd mmm dd yyyy HH:MM:ss",
    shortDate:      "m/d/yy",
    mediumDate:     "mmm d, yyyy",
    longDate:       "mmmm d, yyyy",
    fullDate:       "dddd, mmmm d, yyyy",
    shortTime:      "h:MM TT",
    mediumTime:     "h:MM:ss TT",
    longTime:       "h:MM:ss TT Z",
    isoDate:        "yyyy-mm-dd",
    isoTime:        "HH:MM:ss",
    isoDateTime:    "yyyy-mm-dd'T'HH:MM:ss",
    isoUtcDateTime: "UTC:yyyy-mm-dd'T'HH:MM:ss'Z'"
};
publiwide.utils.dayNames = [
    "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat",
    "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"
];
publiwide.utils.monthNames =
    [
        "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec",
        "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"
    ];

publiwide.utils.dateFormat = function()
{
    var	token = /d{1,4}|m{1,4}|yy(?:yy)?|([HhMsTt])\1?|[LloSZ]|"[^"]*"|'[^']*'/g,
        timezone = /\b(?:[PMCEA][SDP]T|(?:Pacific|Mountain|Central|Eastern|Atlantic) (?:Standard|Daylight|Prevailing) Time|(?:GMT|UTC)(?:[-+]\d{4})?)\b/g,
        timezoneClip = /[^-+\dA-Z]/g,
        pad = function (val, len)
        {
            val = String(val);
            len = len || 2;
            while (val.length < len)
                val = '0' + val;
            return val;
        };

    return function (date, mask, utc)
    {
        if (arguments.length == 1 && Object.prototype.toString.call(date) == "[object String]" && !/\d/.test(date))
        {
            mask = date;
            date = undefined;
        }

        date = date ? new Date(date) : new Date;
        if (isNaN(date))
            throw new SyntaxError('invalid date');

        mask = String(publiwide.utils.dateMasks[mask] || mask || publiwide.utils.dateMasks['default']);

        if (mask.slice(0, 4) == "UTC:")
        {
            mask = mask.slice(4);
            utc = true;
        }

        var	_ = utc ? "getUTC" : "get",
            d = date[_ + "Date"](),
            D = date[_ + "Day"](),
            m = date[_ + "Month"](),
            y = date[_ + "FullYear"](),
            H = date[_ + "Hours"](),
            M = date[_ + "Minutes"](),
            s = date[_ + "Seconds"](),
            L = date[_ + "Milliseconds"](),
            o = utc ? 0 : date.getTimezoneOffset(),
            flags = {
                d:    d,
                dd:   pad(d),
                ddd:  publiwide.utils.dayNames[D],
                dddd: publiwide.utils.dayNames[D + 7],
                m:    m + 1,
                mm:   pad(m + 1),
                mmm:  publiwide.utils.monthNames[m],
                mmmm: publiwide.utils.monthNames[m + 12],
                yy:   String(y).slice(2),
                yyyy: y,
                h:    H % 12 || 12,
                hh:   pad(H % 12 || 12),
                H:    H,
                HH:   pad(H),
                M:    M,
                MM:   pad(M),
                s:    s,
                ss:   pad(s),
                l:    pad(L, 3),
                L:    pad(L > 99 ? Math.round(L / 10) : L),
                t:    H < 12 ? "a"  : "p",
                tt:   H < 12 ? "am" : "pm",
                T:    H < 12 ? "A"  : "P",
                TT:   H < 12 ? "AM" : "PM",
                Z:    utc ? "UTC" : (String(date).match(timezone) || [""]).pop().replace(timezoneClip, ""),
                o:    (o > 0 ? "-" : "+") + pad(Math.floor(Math.abs(o) / 60) * 100 + Math.abs(o) % 60, 4),
                S:    ["th", "st", "nd", "rd"][d % 10 > 3 ? 0 : (d % 100 - d % 10 != 10) * d % 10]
            };

        return mask.replace(token, function ($0)
        {
            return $0 in flags ? flags[$0] : $0.slice(1, $0.length - 1);
        });
    };
}();

publiwide.utils.Dictionary = function()
{
    this.count = function()
    {
        var result = 0;
        for (var x in this)
        {
            if (typeof this[x] != 'function' && this.hasOwnProperty(x))
                result++;
        }
        return result;
    };
    this.remove = function(item)
    {
        for (var x in this)
        {
            if (this.hasOwnProperty(x) && typeof this[x] != 'function' && item.key == this[x].key)
                delete this[x];
        }
    }
    this.toArray = function()
    {
        var result = [];
        for (var x in this)
        {
            if (this.hasOwnProperty(x) && typeof this[x] != 'function')
                result.push(this[x]);
        }
        return result;
    };
};

publiwide.utils.Listeners = function()
{
    this.items = [];
    this.add = function(func)
    {
        this.items.push(func);
    };
    this.remove = function(func)
    {
        var index = this.items.indexOf(func);
        if (index >= 0)
        {
            this.items.splice(index, 1);
            return true;
        }
        return false;
    };
    this.fire = function(data)
    {
        for (var i = 0; i < this.items.length; i++)
            this.items[i](data);
    };
};

Date.prototype.format = function (mask, utc)
{
    return publiwide.utils.dateFormat(this, mask, utc);
};

if (typeof String.prototype.startsWith != 'function')
{
    String.prototype.startsWith = function (str)
    {
        return this.slice(0, str.length) == str;
    };
}
if (typeof String.prototype.endsWith != 'function')
{
    String.prototype.endsWith = function (str)
    {
        return this.slice(-str.length) == str;
    };
}
if (typeof String.prototype.trimLeft != 'function')
{
    String.prototype.trimLeft = function (chars)
    {
        if (chars == null)
            chars = ' ';
        var str = this;
        while (str.length > 0 && chars.indexOf(str.charAt(0)) >= 0)
            str = str.substring(1);
        return str;
    };
}
if (typeof String.prototype.trimRight != 'function')
{
    String.prototype.trimRight = function (chars)
    {
        if (chars == null)
            chars = ' ';
        var str = this;
        while (str.length > 0 && chars.indexOf(str.charAt(str.length - 1)) >= 0)
            str = str.substring(0, str.length - 2);
        return str;
    };
}
if (typeof String.prototype.trimAll != 'function')
{
    String.prototype.trimAll = function (chars)
    {
        var str = this.trimLeft(chars);
        str = str.trimRight(chars);
        return str;
    };
}
if (typeof Element.prototype.offset != 'function')
{
    Element.prototype.offset = function()
    {
        var box = { top: 0, left: 0 };
        var document = viewAnnot.getDocument();
        var window = viewAnnot.getWindow();

        if (typeof this.getBoundingClientRect !== 'undefined')
            box = this.getBoundingClientRect();
        return {
            top: box.top  + (document.scrollTop ? document.scrollTop : window.pageYOffset)  - (document.clientTop ? document.clientTop : 0),
            left: box.left + (document.scrollLeft ? document.scrollLeft: window.pageXOffset) - (document.clientLeft ? document.clientLeft : 0)
        };
    };
}